
/* standard includes */
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <stdarg.h>
#include <fcntl.h>
#include <sys/time.h>


/* X11 includes */
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Cardinals.h>
#include <X11/Xaw/Dialog.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Box.h>
#include <X11/Shell.h>


/* Networking includes */
#include <sys/types.h> 
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

#if 0
#include <arpa/inet.h>
#endif

/* Private includes */
#define IN_UI
#include "atlantis.h"
#include "fatal.h"
#include "undo.h"
#include "ui.h"
#include "ral.h"


static XtAppContext app_context;
static GC gcBlack,gcWhite,gctmp;
static GC gcline,gcwide;
static int width = 300,height = 300;
static XColor playcols[16];
static XColor gray[7];

static int pl,nplay;
static int initializing=1;

char *myname;

#ifdef SMALL
static XPoint hexc[] = {{-8, 5}, {-8,-4}, { 0,-10}, 
			{ 8,-5}, { 8, 4}, { 0, 10}, 
			{-8, 5}};  /* One extra for wrap around */ 

#define HEXX(i,j)  (i*16 + j * 8 + sdx )
#define HEXY(i,j)  (j * (10+5) + sdy)

#define CONVERTXYIJ(x,y,i,j) \
	j = (y+5 - sdy) / 15;   \
	i = ((x+8) - (j * 8 + sdx))/16 
#else

static XPoint hexc[] = {{-14, 8}, {-14,-8}, { 0,-16}, 
			{ 14,-8}, { 14, 8}, { 0, 16}, 
			{-14, 8}};  /* One extra for wrap around */ 

#define HEXX(i,j)  (i*28 + j * 14 + sdx)
#define HEXY(i,j)  (j * (16+8) + sdy)

#define CONVERTXYIJ(x,y,i,j) \
	j = (y+7 - sdy) / 24;   \
	i = ((x+14) - (j * 14 + sdx))/28 

#endif


static int serverfd;
static FILE *serverFILE;

static int brd[MAXB][MAXB];
static int sectors[MAXS][MAXS];
static int hdx,hdy;
static int sdx,sdy;

static int dirsx[7] = { 0,  0,  1,  1,  0, -1, -1};
static int dirsy[7] = { 0, -1, -1,  0,  1,  1,  0};

/* PK: added varying cursor for the field window */
static Cursor hexcursor;
static hexcursoractive = -1;		/* not initialized yet */



int hassunk (int i,int j)
{
  int k,x,y;

  x = SECTX(i,j);
  y = SECTY(i,j);
  for (ALLDIRS(k))
    if (!ISOFFBOARD (SECTX(i,j)+dirsx[k],SECTY(i,j)+dirsy[k]))
      return (0);
  return (1);
}




static void serverprintf (char *str,...)
{
  va_list v;
  char buf[1000];

  va_start (v,str);
  vsprintf (buf,str,v);
  write (serverfd,buf,strlen(buf));
  va_end (v);
}



void calcscreentrans (void)
{
  int minx,maxx,miny,maxy,i,j,x,y;

  minx = miny =  100000;
  maxx = maxy = -100000;
  sdx = sdy = 0;
  for(ALLSECTORS (i,j))
    {
      x = HEXX(SECTX(i,j),SECTY(i,j));
      y = HEXY(SECTX(i,j),SECTY(i,j));
      if (x < minx) minx = x;
      if (y < miny) miny = y;
      if (x > maxx) maxx = x;
      if (y > maxy) maxy = y;
    }
  sdx = 45 - minx;
  sdy = 45 - miny;
  printf ("minx = %d, miny = %d. -> sdx = %d, sdy = %d.\n",minx,miny,sdx,sdy);
  width = maxx - minx + 90;
  height = maxy - miny + 90;
  printf ("maxx = %d, maxy = %d. -> width = %d, height = %d.\n",maxx,maxy,width,height);
}


static void drawtri (int x,int y,int i,GC gc)
{
  XPoint  xpoints[5];

  xpoints[0].x = x + hexc[i].x;
  xpoints[0].y = y + hexc[i].y;
  xpoints[1].x = x + hexc[i+1].x;
  xpoints[1].y = y + hexc[i+1].y;
  xpoints[2].x = x;
  xpoints[2].y = y;
  XFillPolygon (XtDisplay(mw.field),XtWindow (mw.field),gc,
		xpoints,3,Convex,FillSolid);
}


static void drawhex (int i,int j,int bc,int d)
{
  int x,y,k,ntri;
  XPoint xp[10];

  x = HEXX(i,j);
  y = HEXY(i,j);


  for (k=0;k<6;k++) {
    xp[k].x = x + hexc[k].x;
    xp[k].y = y + hexc[k].y;
  }
  if (ISOFFBOARD (i,j) && (bc != 4)) bc = 0;
  if (ISBRON     (i,j)             ) bc = 4;
  XSetForeground(XtDisplay(mw.field), gctmp, gray[bc].pixel);
  XFillPolygon(XtDisplay(mw.field),XtWindow (mw.field),gctmp,
	       xp,6,Convex,FillSolid);

  XSetForeground(XtDisplay(mw.field), gctmp, playcols[PL(i,j)].pixel);
  ntri = NP(i,j);
  if (ntri > 6) ntri = 6;
  switch (ntri) {
  case 6:drawtri (x,y,1,gctmp);
    drawtri (x,y,4,gctmp);
  case 4:drawtri (x,y,5,gctmp);
    drawtri (x,y,2,gctmp);
  case 2:drawtri (x,y,3,gctmp);
    drawtri (x,y,0,gctmp);
    break;
  case 5:drawtri (x,y,3,gctmp);
    drawtri (x,y,5,gctmp);
  case 3:drawtri (x,y,2,gctmp);
    drawtri (x,y,4,gctmp);
  case 1:drawtri (x,y,0,gctmp);
  case 0:break;
  default:
    printf ("Got a problem: board at %d,%d contains %d stones...\n",i,j,ntri);
  }
  if (ISOFFBOARD(i,j)) {
    XSetForeground (XtDisplay(mw.field), gcline, gray[4].pixel);
    XSetForeground (XtDisplay(mw.field), gcwide, gray[4].pixel);
  } else {
    XSetForeground (XtDisplay(mw.field), gcline, gray[0].pixel);
    XSetForeground (XtDisplay(mw.field), gcwide, gray[0].pixel);
  }
  for (k=d+2;k<d+5;k++) {
    XDrawLine (XtDisplay(mw.field),XtWindow (mw.field),
	       gcline, 
	       x+hexc[(k  )%6].x,y+hexc[(k  )%6].y,
	       x+hexc[(k+1)%6].x,y+hexc[(k+1)%6].y);
  }
  for (k=d+5;k<d+8;k++) {
    XDrawLine (XtDisplay(mw.field),XtWindow (mw.field),
	       d?gcwide:gcline, 
	       x+hexc[(k  )%6].x,y+hexc[(k  )%6].y,
	       x+hexc[(k+1)%6].x,y+hexc[(k+1)%6].y);
  }
}

void releasehexcursor(void)
{
  /* De-allocate cursor (I /think/ it takes up memory in the X-server) */
  if(hexcursoractive == 1) {
    XUndefineCursor(XtDisplay(mw.field), XtWindow(mw.field));
    XFreeCursor(XtDisplay(mw.field), hexcursor);
    hexcursoractive = 0;
  }
}


static void hexcursordrawtri (int i, Pixmap bitmap, GC gc)
{
  XPoint  xpoints[10];
  
  xpoints[0].x = 15;
  xpoints[0].y = 15;
  xpoints[1].x = 15 + hexc[i].x;
  xpoints[1].y = 15 + hexc[i].y;
  xpoints[2].x = 15 + hexc[i+1].x;
  xpoints[2].y = 15 + hexc[i+1].y;
  xpoints[3].x = 15;
  xpoints[3].y = 15;
  XDrawLines (XtDisplay(mw.field), bitmap, gc, xpoints, 4, CoordModeOrigin);
  /*
   * Increase the thickness of the "outer" edge of the triangle.
   */
  if(xpoints[1].x > 15)
    xpoints[1].x--;
  else if(xpoints[1].x < 15)
    xpoints[1].x++;
  if(xpoints[1].y > 15)
    xpoints[1].y--;
  else if(xpoints[1].y < 15)
    xpoints[1].y++;
  if(xpoints[2].x > 15)
    xpoints[2].x--;
  else if(xpoints[2].x < 15)
    xpoints[2].x++;
  if(xpoints[2].y > 15)
    xpoints[2].y--;
  else if(xpoints[2].y < 15)
    xpoints[2].y++;
  XDrawLines (XtDisplay(mw.field), bitmap, gc, &xpoints[1], 2,
	      CoordModeOrigin);
}


static void hexcursordrawhex (int ntri)
{
  XColor dummyColor;
  XPoint  xpoints[10];
  int k;
  static Pixmap hexcursorforeground;
  static Pixmap hexcursormask;
  static GC cursorGC;
  static XGCValues cursorgcv;
  static XColor cursorbgcolor;

  releasehexcursor();

  if(hexcursoractive == -1) {
    /*
     * PK: Allocate stuff for the variable cursor, clear the foreground
     * and set (create ?) the colors.
     */
    hexcursorforeground = XCreatePixmap(XtDisplay(mw.field),
					RootWindow(XtDisplay(mw.field),
						   DefaultScreen(XtDisplay(mw.field))), 32, 32, 1);
    hexcursormask = XCreatePixmap(XtDisplay(mw.field),
				  RootWindow(XtDisplay(mw.field),
					     DefaultScreen(XtDisplay(mw.field))), 32, 32, 1);
    cursorGC = XCreateGC(XtDisplay(mw.field), hexcursorforeground,
			 GCForeground | GCBackground, &cursorgcv);
    XSetFunction(XtDisplay(mw.field), cursorGC, GXclear);
    XFillRectangle(XtDisplay(mw.field), hexcursorforeground, cursorGC,
		   0, 0, 32, 32);
    XAllocNamedColor (XtDisplay (mw.field),
		      DefaultColormap (XtDisplay (mw.field), 0), "black",
		      &cursorbgcolor, &dummyColor);
  } 

  /* Clear the cursormask */
  XSetFunction(XtDisplay(mw.field), cursorGC, GXclear);
  XFillRectangle(XtDisplay(mw.field), hexcursormask, cursorGC, 0, 0, 32, 32);

  /* Prepare to draw in the cursormask */
  XSetFunction(XtDisplay(mw.field), cursorGC, GXset);

  switch (ntri) {
  case 4: hexcursordrawtri (5, hexcursormask, cursorGC);
    hexcursordrawtri (2, hexcursormask, cursorGC);
  case 2: hexcursordrawtri (3, hexcursormask, cursorGC);
    hexcursordrawtri (0, hexcursormask, cursorGC);
    break;
  case 5: hexcursordrawtri (3, hexcursormask, cursorGC);
    hexcursordrawtri (5, hexcursormask, cursorGC);
  case 3: hexcursordrawtri (2, hexcursormask, cursorGC);
    hexcursordrawtri (4, hexcursormask, cursorGC);
  case 1: hexcursordrawtri (0, hexcursormask, cursorGC);
    break;
  default:
    printf ("Got a problem: Can't create cursor for %d stones...\n",
	    ntri);
    return;			/* Don't create a blank cursor */
  }
  /*
   * Draw a hexagon of the size of a hex in the mask and in the foreground.
   */
  for(ALLDIRS(k)) {
    xpoints[k].x = 15 + hexc[k].x;
    xpoints[k].y = 15 + hexc[k].y;
  }
  XDrawLines (XtDisplay(mw.field), hexcursorforeground, cursorGC,
	      xpoints, 7, CoordModeOrigin);
  XDrawLines (XtDisplay(mw.field), hexcursormask, cursorGC,
	      xpoints, 7, CoordModeOrigin);

  /* Create the cursor */
  hexcursor = XCreatePixmapCursor(XtDisplay(mw.field), hexcursorforeground,
				  hexcursormask, &playcols[pl], &cursorbgcolor, 15, 15);
  /* Activate the cursor */
  XDefineCursor(XtDisplay(mw.field), XtWindow(mw.field), hexcursor);
  hexcursoractive = 1;
}


static void draw_a_sector (int i,int j,int drawit)
{
  int k,bc;

  bc = drawit? 1 + ( (i + j*2) % 3) : 4;
  for (ALLDIRS(k))
    drawhex (SECTX(i,j)+dirsx[k],SECTY(i,j)+dirsy[k],bc,k);
}



void doredraw (void)
{
  int i,j,k,bc;

  /* XFillRectangle (XtDisplay(mw.field),XtWindow (mw.field),gcWhite,0,0,width,height);  */
  XSetForeground(XtDisplay(mw.field), gctmp, playcols[pl].pixel);
  XFillRectangle (XtDisplay(mw.field),XtWindow (mw.field),gctmp,10,10,20,20);
  for(ALLSECTORS(i,j)) {
    bc = 1 + ( (i + j*2) % 3);
    if (!hassunk (i,j))
      for (ALLDIRS(k))
	drawhex (SECTX(i,j)+dirsx[k],SECTY(i,j)+dirsy[k],bc,k);
  }
}


static void expose (w, event, params, nparams)
    		Widget          w;
		XEvent         *event;
		String         *params;
		Cardinal       *nparams;
{
  Arg args[]= {  {XtNwidth,0}, {XtNheight,0}};

  if (initializing)
    return;

  /* Have these become pointers???? */
  args[0].value = (int) &width;
  args[1].value = (int) &height;

  XtGetValues(w, args, 2);

  if (event->xexpose.count != 0) return;
  printf("New width and height are: %d and %d.\n",width,height);
  doredraw ();
}

static void motion (Widget w,XEvent * event, String params, Cardinal nparams)
{

}

static void simplebutton (Widget w, XtPointer clientdata)
{
  serverprintf ("%s\n",clientdata);
}


static void keys (Widget w,XEvent * event, String params, Cardinal nparams)
{
  KeySym ch;

  XtTranslateKeycode (XtDisplay (w),event->xkey.keycode,event->xkey.state,&event->xkey.state,&ch);
  switch (ch) {
  case 'q': case 'Q': simplebutton (w,"quit"); break;
  case 'u': case 'U': simplebutton (w,"undo"); break;
  case 'r': case 'R': simplebutton (w,"redo"); break;
  case 'd': case 'D': simplebutton (w,"done"); break;
  }
}


static void buttons (w, event, params, nparams)
    		Widget          w;
		XEvent         *event;
		String         *params;
		Cardinal       *nparams;
{
  int x,y,i,j;

  x = event->xbutton.x;
  y = event->xbutton.y;
  CONVERTXYIJ(x,y,i,j);
  switch (event->xbutton.button) {
  case 1:
    serverprintf ("source %d %d\n",i,j);
    break;
  case 2:
    serverprintf ("dest %d %d\n",i,j);
    break;
  case 3:
    simplebutton (w,resources.button3);
    break;
  }
}



#if 0
void load (Widget w, XtPointer clientdata)
{
  doload ("atlantis.save");
}

void save (Widget w, XtPointer clientdata)
{
  dosave ("atlantis.save");
}
#endif

static void debug (Widget w, XtPointer clientdata)
{
  int i,j;

  for (i=0;i<MAXB;i++)
    printf ("%02d",i);
  putchar ('\n');
  for (j=0;j<MAXB;j++) {
    for (i=0;i<MAXB;i++)
      if      (ISOFFBOARD(i,j)) printf ("  ");
      else if (ISEMPTY(i,j))    printf ("..");
      else printf ("%d%d",NP(i,j),PL(i,j));
    putchar ('\n');
  }
}


static void resync (Widget w, XtPointer clientdata)
{
  serverprintf ("getboard\n");
}


static int connect_server (char *remotehostname,char *hostname,int port)
{
  struct hostent *remotehostdata;
  struct sockaddr_in remoteaddress;
   
  if (!remotehostname) remotehostname = hostname;

  if ((remotehostdata = gethostbyname(remotehostname)) == NULL) 
    fatal ("%s: gethostbyname", myname);

  memset (&remoteaddress, '\0', sizeof (remoteaddress));

  remoteaddress.sin_family = AF_INET;
  remoteaddress.sin_port = htons(port);
  remoteaddress.sin_addr.s_addr = *(u_long *) remotehostdata->h_addr;

  if ((serverfd = socket(AF_INET,SOCK_STREAM,0)) < 0)
    fatal ("%s: socket", myname);

  if(connect(serverfd, (struct sockaddr *)&remoteaddress, sizeof (remoteaddress)) < 0)
    fatal ("%s: connect", myname);

  if (fcntl (serverfd,F_SETFL,O_NONBLOCK) < 0)
    fatal ("%s: fcntl", myname);

  return (serverfd);
  
}


static XrmOptionDescRec options[] = {
  { "-np", ".nplayers",  XrmoptionSepArg, NULL },
  { "-si", ".size",      XrmoptionSepArg, NULL }, 
  { "-bo", ".board",     XrmoptionSepArg, NULL }, 
  { "-pln", ".pln",      XrmoptionSepArg, NULL }, 
  { "-server",".server", XrmoptionSepArg, NULL }, 
  { "-port", ".port",    XrmoptionSepArg, NULL }, 
  { "-user", ".user",    XrmoptionSepArg, NULL }, 
  { "-button3", ".button3",XrmoptionSepArg, NULL }, 
};



static XtResource resource_list[] = {
  { "nplayers", "Nplayers", XtRCardinal, sizeof(int),
      XtOffsetOf(Resources, nplayers), XtRImmediate, (XtPointer) 6 },
  { "server",   "Server",   XtRString, sizeof(int),
      XtOffsetOf(Resources, server),   XtRImmediate, (XtPointer) "localhost" },
  { "port",      "Port",    XtRCardinal, sizeof(int),
      XtOffsetOf(Resources, port),     XtRImmediate, (XtPointer) 10000 },
  { "pln",      "Pln",      XtRCardinal, sizeof(int),
      XtOffsetOf(Resources, pln),     XtRImmediate, (XtPointer) -2 },
  { "size",     "Size",     XtRCardinal, sizeof(int),
      XtOffsetOf(Resources, size),     XtRImmediate, (XtPointer) 5 },
  { "board",    "Board",    XtRString, sizeof(int),
      XtOffsetOf(Resources, board),    XtRImmediate, (XtPointer) NULL },
  { "user",     "User",     XtRString, sizeof(int),
      XtOffsetOf(Resources, user),     XtRImmediate, (XtPointer) NULL },
  { "color0",   "Color0",   XtRString, sizeof(int),
      XtOffsetOf(Resources, colornames[0]),XtRImmediate, (XtPointer) "red" },
  { "color1",   "Color1",   XtRString, sizeof(int),
      XtOffsetOf(Resources, colornames[1]),XtRImmediate, (XtPointer) "green" },
  { "color2",   "Color2",   XtRString, sizeof(int),
      XtOffsetOf(Resources, colornames[2]),XtRImmediate, (XtPointer) "blue" },
  { "color3",   "Color3",   XtRString, sizeof(int),
      XtOffsetOf(Resources, colornames[3]),XtRImmediate, (XtPointer) "magenta" },
  { "color4",   "Color4",   XtRString, sizeof(int),
      XtOffsetOf(Resources, colornames[4]),XtRImmediate, (XtPointer) "cyan" },
  { "color5",   "Color5",   XtRString, sizeof(int),
      XtOffsetOf(Resources, colornames[5]),XtRImmediate, (XtPointer) "yellow" },
  { "color6",   "Color6",   XtRString, sizeof(int),
      XtOffsetOf(Resources, colornames[6]),XtRImmediate, (XtPointer) "pink" },
  { "color7",   "Color7",   XtRString, sizeof(int),
      XtOffsetOf(Resources, colornames[7]),XtRImmediate, (XtPointer) "brown" },
  { "gray0",    "gray0",    XtRString, sizeof(int),
      XtOffsetOf(Resources, graynames[0]), XtRImmediate, (XtPointer)"gray0"},
  { "gray1",    "gray1",    XtRString, sizeof(int),
      XtOffsetOf(Resources, graynames[1]), XtRImmediate, (XtPointer)"gray70"},
  { "gray2",    "gray2",    XtRString, sizeof(int),
      XtOffsetOf(Resources, graynames[2]), XtRImmediate, (XtPointer)"gray77"},
  { "gray3",    "gray3",    XtRString, sizeof(int),
      XtOffsetOf(Resources, graynames[3]), XtRImmediate, (XtPointer)"gray85"},
  { "gray4",    "gray4",    XtRString, sizeof(int),
      XtOffsetOf(Resources, graynames[4]), XtRImmediate, (XtPointer)"gray100"},
  { "button3",    "button3",    XtRString, sizeof(int),
      XtOffsetOf(Resources, button3), XtRImmediate, (XtPointer)"done"},
};


Resources resources;
static MainWindowRec mw;
static XtAppContext app_context;


static Arg shell_args[] = {
  { XtNtitle, (XtArgVal) "New atlantis Client" },
};


static String fallback_resources[] = {
  NULL,
};


static Arg formargs[] = {
  {XtNwidth,400},
  {XtNheight,400},
};

static Arg fieldargs[] = {
  {XtNfromVert,0},
  {XtNwidth,0},
  {XtNheight,0},
};

static Arg mbargs[] = {
  {XtNlabel,(XtArgVal)
   "                                                       "},
};




static XtActionsRec Actions[] = {
  {"buttons",      (XtActionProc) buttons},
  {"expose",       (XtActionProc) expose},
  {"motion",       (XtActionProc) motion},
  {"keys",         (XtActionProc) keys},
};

static char    *overrideTranslations =
"<Expose>:              expose()\n"
"<ResReq>:              expose()\n"
"<KeyPress>:            keys()\n"
"<ButtonPress>:         buttons()";


static void newinput (Widget w, XtPointer clientdata)
{
  char  *bp;
  int code;
  int sx,sy,x,y,k;	
  char s[10][10];
  char ch,ct;
  char playername[80];

  /* printf ("newinput called fd = %d.\n",serverfd); */
  do {
    bp = read_a_line (serverfd);
    if (ral_eof (serverfd)) {
	printf ("Shit. Server closed connection.\n");
	exit (1);
      }
    if (bp == NULL)
      return;
    printf ("Got '%s'.\n",bp);
    sscanf (bp,"%d",&code);
    if (!initializing) {
	mbargs[0].value = (XtArgVal) bp;
	XtSetValues (mw.menubar,mbargs,1);
      }
    switch (code) {
    case 100:  /* Welcome. Now register my username. */
      if (resources.pln == -2)
	serverprintf ("user %s\n",resources.user);
      else
	serverprintf ("user %s %d\n",resources.user,resources.pln);
      break;
    case 120:
      sscanf (bp,"%d- %d %d %d %d",&code,&nplay,&pl,&hdx,&hdy);
      break;
    case 129:
      sscanf (bp,"%d%c %d %d",
	      &code,&ct,&sx,&sy);
      x = SECTX (sx,sy);
      y = SECTY (sx,sy);
      for (ALLDIRS(k))
	MAKEKRATER (x+dirsx[k],y+dirsy[k]);
      draw_a_sector (sx,sy,FALSE);
      break;
    case 180: /* Server tells us goodbye */
      releasehexcursor();
      exit (0);
      break;
    case 130: 
    case 121:
      releasehexcursor();
      sscanf (bp,"%d%c %d %d %s %s %s %s %s %s %s",
	      &code,&ct,&sx,&sy,s[0],s[1],s[2],s[3],s[4],s[5],s[6]);
      x = SECTX(sx,sy);
      y = SECTY(sx,sy);
      sectors[sx][sy]++;
      for (ALLDIRS(k))
	switch (ch = s[k][0]) {
	case CRATER:
	  MAKEKRATER (x+dirsx[k],y+dirsy[k]);break;
	case EMPTY:
	  SETONBOARD (x+dirsx[k],y+dirsy[k]);
	  MAKEEMPTY  (x+dirsx[k],y+dirsy[k]);break;
	case SOURCE:
	case PLAYER:
	  SETONBOARD (x+dirsx[k],y+dirsy[k]);
	  SETPLAYER  (x+dirsx[k],y+dirsy[k],s[k][1] - '0');
	  SETPIECES  (x+dirsx[k],y+dirsy[k],s[k][2] - '0');
	  if (ch == SOURCE)
	    SETBRON (x+dirsx[k],y+dirsy[k]);
	  break;
	default:
	  printf ("Don't understand board change from server '%s'\n",bp);
	  break;
	}
      if (!initializing) 
	draw_a_sector (sx,sy,TRUE);
      break;
    case 122: /* That's it. */
      if (initializing)
	{
	  calcscreentrans ();
	  initializing = 0;
	}
      else
	printf ("Getboard used while not initializing?\n");
      break;
    case 135:
      releasehexcursor();
      sscanf (bp,"%d%c %d %s",&code,&ct,&pl,playername);
      XSetForeground(XtDisplay(mw.field), gctmp, playcols[pl].pixel);
      XFillRectangle (XtDisplay(mw.field),XtWindow (mw.field),gctmp,
		      10,10,20,20);
      XFillRectangle (XtDisplay(mw.field),XtWindow (mw.field),gcWhite,
		      30,10,60,20);
      XDrawString (XtDisplay(mw.field),XtWindow (mw.field),gcBlack,
		   35,20,playername,strlen(playername));
      break;
    case 140:
    case 141:
      break;
    case 142:
      printf ("Its my turn!!! Go ahead and play...\n");
      break;
    case 160: /* Source ok.... */
      sscanf (bp, "%d%c %d", &code, &ct, &k);
      hexcursordrawhex (k);
      break;
    case 162: /* Name OK, now get the board. */
      serverprintf ("getboard\n");
      break;
    case 430: /* Problem with playernumber. */
    case 441: /* Name ... already taken. */
      {
	static int t=0;

	t++;	
	serverprintf ("user %s_%d\n",resources.user,t);
      }
      break;
    case 400:
    case 460:
      printf ("Mhhhh should have kept my mouth shut....\n");
      break;
    default:
      releasehexcursor();
      printf ("Unknown Server reply code....\n");
      break;	
    }
  } while (more_input_available (serverfd));
}



static void CvtStringToCardinal(XrmValue *args, Cardinal *n_args,
                                XrmValue *fromVal, XrmValue *toVal)
{
  static int d;

  sscanf (fromVal->addr,"%d",&d);
  toVal->addr = (void *)&d;
  toVal->size = sizeof(d);
}


void main (int argc,char **argv)
{
  XColor          srcColor, dummyColor;
  XtTranslations  newTranslations;
  int i;
  int serverfd;
  int id;

  myname = argv[0];
  mw.shell = XtAppInitialize(&app_context, "Atlantis", options, XtNumber(options),
                             &argc, argv, fallback_resources, shell_args,
                             XtNumber(shell_args) );

  XtAppAddConverter(app_context, XtRString, XtRCardinal, 
                    CvtStringToCardinal, NULL, ZERO);


  XtGetApplicationResources(mw.shell, &resources, resource_list,
                            XtNumber(resource_list), NULL, ZERO);

  serverfd = connect_server (resources.server,resources.server,resources.port);
  serverFILE = fdopen (serverfd,"rw");
  if (!resources.user)
    resources.user = getenv ("LOGNAME");

  while (initializing)
    {
      fd_set infds;
	
      FD_ZERO (&infds);
      FD_SET (serverfd,&infds);
      select (serverfd+1,&infds,NULL,NULL,NULL);
      newinput (NULL,NULL);
    }

  id =  XtAppAddInput(app_context, serverfd, (XtPointer)XtInputReadMask,
		      (XtInputCallbackProc)newinput, (XtPointer)serverfd);

  mw.form = XtCreateManagedWidget("form", formWidgetClass, mw.shell,
				  formargs,XtNumber(formargs));

  mw.menubar = XtCreateManagedWidget("menubar", dialogWidgetClass, mw.form,
				     mbargs,XtNumber(mbargs));

  XawDialogAddButton (mw.menubar, "quit",(XtCallbackProc) simplebutton,  "quit");
#if 0
  XawDialogAddButton (mw.menubar, "load",(XtCallbackProc) load,  (XtPointer) NULL);
  XawDialogAddButton (mw.menubar, "save",(XtCallbackProc) save,  (XtPointer) NULL);
#endif
  XawDialogAddButton (mw.menubar, "undo",(XtCallbackProc)simplebutton, "undo");
  XawDialogAddButton (mw.menubar, "redo",(XtCallbackProc)simplebutton, "redo");
  XawDialogAddButton (mw.menubar, "done",(XtCallbackProc)simplebutton, "done");
  XawDialogAddButton (mw.menubar, "debug",(XtCallbackProc)debug, (XtPointer) NULL);
  XawDialogAddButton (mw.menubar, "resync",(XtCallbackProc)resync, (XtPointer) NULL);

  fieldargs[0].value = (XtArgVal) mw.menubar;
  fieldargs[1].value = (XtArgVal) width;
  fieldargs[2].value = (XtArgVal) height; 

  mw.field = XtCreateManagedWidget("field", formWidgetClass, mw.form, 
				   fieldargs,XtNumber (fieldargs));

  XtAppAddActions(app_context,Actions, XtNumber(Actions));
  newTranslations = XtParseTranslationTable(overrideTranslations);
  XtOverrideTranslations(mw.field, newTranslations);

  XtRealizeWidget(mw.shell);


  gcWhite = XCreateGC(XtDisplay(mw.field), XtWindow(mw.field), (unsigned long) 0, NULL);
  XAllocNamedColor(XtDisplay(mw.field), DefaultColormap(XtDisplay(mw.field), 0), "white" , &srcColor, &dummyColor);
  XSetForeground(XtDisplay(mw.field), gcWhite, srcColor.pixel);

  gcBlack = XCreateGC(XtDisplay(mw.field), XtWindow(mw.field), (unsigned long) 0, NULL);
  XAllocNamedColor(XtDisplay(mw.field), DefaultColormap(XtDisplay(mw.field), 0), "black" , &srcColor, &dummyColor);
  XSetForeground(XtDisplay(mw.field), gcBlack, srcColor.pixel);

  gcline = XCreateGC(XtDisplay(mw.field), XtWindow(mw.field), (unsigned long) 0, NULL);
  gcwide = XCreateGC(XtDisplay(mw.field), XtWindow(mw.field), (unsigned long) 0, NULL);
  XSetLineAttributes (XtDisplay(mw.field),gcwide, 2,
		      LineSolid,CapNotLast,JoinRound);

  gctmp = XCreateGC(XtDisplay(mw.field), XtWindow(mw.field), (unsigned long) 0, NULL);
  XSetForeground(XtDisplay(mw.field), gctmp, srcColor.pixel);


  for (i=0;i<sizeof(resources.colornames)/sizeof (char *);i++)
    XAllocNamedColor(XtDisplay(mw.field), DefaultColormap(XtDisplay(mw.field), 0), 
		     resources.colornames[i], &playcols[i], &dummyColor);

  for (i=0;i<sizeof(resources.graynames)/sizeof (char *);i++)
    XAllocNamedColor(XtDisplay(mw.field), DefaultColormap(XtDisplay(mw.field), 0), 
		     resources.graynames[i], &gray[i], &dummyColor);

  XtAppMainLoop (app_context);
}



