/*
 * atlantisd.c 
 * Written by Roger E. Wolff.
 */


#include <stdio.h>
#include <stdlib.h>

#include "atlantis.h"
#include "ui.h"
#include "undo.h"

#include <sys/types.h>

#include "serve.h"

int dirsx[7] = { 0,  0,  1,  1,  0, -1, -1};
int dirsy[7] = { 0, -1, -1,  0,  1,  1,  0};

int brd[MAXB][MAXB];
int sectors[MAXS][MAXS];
int nplay;
int pl;
int hdx,hdy;
int score[MAXPL+2];

extern int width,height;
extern int maxused_x, maxused_y;      /* PK */

    
int hassunk (int i, int j)
{
  int k,x,y;

  x = SECTX(i,j);
  y = SECTY(i,j);
  for (ALLDIRS(k))
    if (!ISOFFBOARD (SECTX(i,j)+dirsx[k], SECTY(i,j)+dirsy[k]))
      return (0);
  return (1);
}


void clearmodified (int i, int j)
{
  int k,x,y;

  x = SECTX(i,j);
  y = SECTY(i,j);
  for (ALLDIRS(k))
    CLRMODIFIED (x+dirsx[k], y+dirsy[k]);
}


int hasmodified (int i, int j)
{
  int k,x,y;

  x = SECTX(i,j);
  y = SECTY(i,j);
  for (ALLDIRS(k))
    if (ISMODIFIED(x+dirsx[k],y+dirsy[k]))
      return (1);
  return (0);
}


int nadjacent (int x,int y)
{
  int i,n;

  for (i=1,n=0;i<7;i++) {
    if (ISBOARD (x+dirsx[i],y+dirsy[i]))
      n++; 
  }
  /*printf ("Nadjacent (%d,%d) = %d.\n",x,y,n);*/
  return n;
}


void doplof (int x,int y)
{
  int i,pl,xx,yy;

  if (nadjacent (x,y) < NP (x,y))
    printf ("%d overkill at (%d,%d).\n",NP(x,y) - nadjacent (x,y),x,y);
  pl = PL(x,y);
  for (i=1;i<7;i++)
    {
      xx = x+dirsx[i];
      yy = y+dirsy[i];
      if (ISBOARD (xx,yy)) {
	if (ISEMPTY(xx,yy))
	  SETPLPI(xx,yy,pl,1);
	else {
	  if (PL(xx,yy) == pl)
	    SETPIECES (xx,yy,NP(xx,yy)+1);
	  else
	    SETPIECES (xx,yy,NP(xx,yy)-1);
	}
	SETMODIFIED (xx,yy);
      }
    }
  if (ISBRON (x,y))
    MAKEKRATER (x,y);
  else
    MAKEBRON (x,y);
  SETMODIFIED (x,y); 
  /* dopredraw (); usleep (200000); */
}


void checkplof (int p)
{
  int i,j,flag;

  flag = 1;
  while (flag) {
    flag = 0;
    for (i=0;i<MAXB;i++)
      for (j=0;j<MAXB;j++)
	if (!ISOFFBOARD(i,j) &&
	    (PL(i,j) == p) &&
	    (NP(i,j) >= nadjacent (i,j))) {
	  doplof (i,j);
	  flag++;
	}
  }
}


void dobron (int p)
{
  int i,j;

  for (i=0;i<MAXB;i++)
    for (j=0;j<MAXB;j++)
      if (!ISOFFBOARD(i,j) &&
	  (PL(i,j) == p) &&
	  (ISBRON(i,j))) {
	SETPIECES (i,j,NP(i,j)+1);
	SETMODIFIED (i,j);
      }
}


void setsector (int i,int j,int pl,int pi)
{
  int x,y,k;

  x = SECTX(i,j);
  y = SECTY(i,j);
  for (ALLDIRS(k)) {
    SETPLPI (x+dirsx[k],y+dirsy[k],pl,pi);
    SETONBOARD(x+dirsx[k],y+dirsy[k]);
  }
}


void clearsector (int i,int j)
{
  setsector (i,j,0,0);
}


void calchextrans (void)
{
  int minx, maxx, miny, maxy, i, j, x, y;

  minx = miny =  100000;
  maxx = maxy = -100000;
  hdx = 0;
  hdy = 0;
  for(ALLSECTORS (i,j)) {
    x = SECTX(i,j);
    y = SECTY(i,j);
    if (x < minx) minx = x;
    if (y < miny) miny = y;
    if (x > maxx) maxx = x;
    if (y > maxy) maxy = y;
  }
  hdx = 2 - minx;
  hdy = 2 - miny;
  printf ("minx = %d, miny = %d. -> hdx = %d, hdy = %d.\n",minx,miny,hdx,hdy);
}


void makestandard (int sz)
{
  int i, j;

  for (j=0;j<sz*2-1;j++) 
    for (i = max(0,sz-1-j);i < min (3*sz-2-j,2*sz-1);i++)
      sectors[i][j] = -1;

  sectors[sz-1][0] = 1;
  switch (nplay) {
  case 2: 
    sectors [  sz-1][2*sz-2] = 2;
    break;
  case 3: 
    sectors [2*sz-2][  sz-1] = 2;
    sectors [     0][2*sz-2] = 3;
    break;
  case 4:
    sectors [2*sz-2][     0] = 2;
    sectors [  sz-1][2*sz-2] = 3;
    sectors [     0][2*sz-2] = 4;
    break;
  case 6:
    sectors [     0][  sz-1] = 6;
  case 5:
    sectors [2*sz-2][     0] = 2;
    sectors [2*sz-2][  sz-1] = 3;
    sectors [  sz-1][2*sz-2] = 4;
    sectors [     0][2*sz-2] = 5;
    break;
  default:
    printf ("Invalid nplayers (%d) in makestandard.\n",nplay);
    exit (1);
  }
} 


void readboard (FILE *fp)
{
  int i=1,j,k,ch,pl;

  nplay = 0;
  j = 0;
  for (ch = '\n';ch != EOF;ch = getc (fp))
    {
      switch (ch) {
      case ' ':i++;break;
      case '*':i++; sectors[i/2][j] = -1;break;
      case '0':
      case '1':
      case '2':
      case '3':
      case '4':
      case '5':i++; 
	pl = ch - '0';
	sectors[i/2][j] = pl + 1;
	if (pl >= nplay) 
	  nplay = pl+1;
	break;
      case '\n':
	j++;
	if (j & 1) {
	  for (k=20;k > 0;k--)
	    for (i=0;i<j;i++)
	      sectors[k][i] = sectors[k-1][i];
	}
	i=1;
	break;
      default: printf ("Illegal character in file: '%c' (%d).\n",(char)ch,(int)ch);
	exit (1);
      }
    }
}


void clearboard (void)
{
  int i,j;

  for (i=0;i<MAXB;i++)
    for (j=0;j<MAXB;j++)
      MAKEOFFBOARD (i,j);
  for (i=0;i<MAXS;i++)
    for (j=0;j<MAXS;j++)
      sectors[i][j] = 0;
}


void initboard (int np,int sz,FILE *fp)
{
  int i,j;
  int x,y;

  nplay = np;
  if ((nplay < 2) || (nplay > 6)) {
    printf ("Can't play atlantis with %d players.\n",nplay);
    exit (1);
  }
  clearboard ();
  if (fp == NULL)
    makestandard (sz);
  else 
    readboard (fp);
  calchextrans ();
 
  for(ALLSECTORS (i,j))
    if (sectors[i][j] < 0)
      clearsector (i,j);
    else
      setsector (i,j,sectors[i][j]-1,1);
 
  for(ALLHEXS(x,y))		/* PK: find used boardarray area */
    if(!ISOFFBOARD(x,y)) {
      if(x >= maxused_x)
	maxused_x = x + 1;
      if(y >= maxused_y)
	maxused_y = y + 1;
    }
}


void printadj (void)
{
  int i,j;

  for (i=0;i<MAXB;i++)
    printf ("%02d",i);
  putchar ('\n');
  for (j=0;j<MAXB;j++) {
    for (i=0;i<MAXB;i++)
      if      (ISOFFBOARD(i,j)) printf ("  ");
      else printf ("%d ",nadjacent (i,j));
    putchar ('\n');
  }
}


void printboard (void)
{
  int i,j;

  for (i=0;i<MAXB;i++)
    printf ("%02d",i);
  putchar ('\n');
  for (j=0;j<MAXB;j++) {
    for (i=0;i<MAXB;i++)
      if      (ISOFFBOARD(i,j)) printf ("  ");
      else if (ISEMPTY(i,j))    printf ("..");
      else printf ("%d%d",NP(i,j),PL(i,j));
    putchar ('\n');
  }
}


void init_moves (int pl)
{
  int i,j;

  for (i=0;i<MAXB;i++)
    for (j=0;j<MAXB;j++)
      if (ISBOARD (i,j))
	if (ISEMPTY (i,j) || (PL (i,j) != pl))
	  SETNPP(i,j,0);
	else
	  SETNPP(i,j,NP(i,j));
}


void findsector (int x,int y,int *ii,int *jj)
{
  int i,j,k;

  for(ALLSECTORS(i,j))
    for (ALLDIRS(k))
      if ( (x == SECTX(i,j)+dirsx[k]) &&
	   (y == SECTY(i,j)+dirsy[k])) {
	*ii = i;
	*jj = j;
	return;
      }
  *ii = -1;
  *jj = -1;
}


int convert(int x0,int y0,int x1,int y1,int *nn,int *dd)
{
  int dx,dy;
  int n,d;

  dx = x1 - x0;
  dy = y1 - y0;

  n = abs (dx);
  if (n == 0) n = abs (dy);
  if (n == 0) return 0;

  for (d=1;d<7;d++)
    if ((dx == n * dirsx[d]) &&
        (dy == n * dirsy[d])) {
      *nn = n;
      *dd = d;
      return 1;
    }
  *dd = 0;
  return (0);
}


int verify_move (int x,int y,int n,int d)
{
  int i;

  if (d == 0) { 
    printf ("Not straight\n");
    return (0);
  }
  if (pl != PL(x,y)) {
    printf ("Not your stone.\n");
    return (0);
  }
  if (n  >  NPP(x,y)) {
    printf ("Not that many stones.\n");
    return (0);
  }
  for (i=0;i<=n;i++)
    if (!ISBOARD (x+dirsx[d]*i,y+dirsy[d]*i)) {
      printf ("Can't jump over an offboard pos.\n");
      return (0);
    }
  return 1;
}


void domove (int x,int y,int n,int d)
{
  int i,j,k,nleft,pl;

  nleft = n;
  pl = PL(x,y);
  SETPIECES(x,y,NP(x,y)-n);
  if (NP(x,y) == 0)
    MAKEEMPTY (x,y);
  SETMODIFIED (x,y);
  findsector (x,y,&i,&j);
  for (ALLDIRS(k))
    SETNPP(SECTX(i,j)+dirsx[k],SECTY(i,j)+dirsy[k],0);

  for (i=0;i<n;i++) {
    x += dirsx[d];
    y += dirsy[d];
    if ((!ISEMPTY(x,y)) && (PL(x,y) != pl)) {
      if (nleft >= NP (x,y)) {
	nleft -= NP (x,y);
	MAKEEMPTY (x,y);
      } else {
	SETPIECES(x,y,NP(x,y)-nleft);
	nleft = 0;
      }
      SETMODIFIED (x,y);
    }
  }
  if (nleft > 0)
    SETPLPI(x,y,pl,NP(x,y) + nleft);
  SETMODIFIED (x,y);
}


int hasone (int i,int j,int pl)
{
  int x,y;

  x = SECTX (i,j);
  y = SECTY (i,j);
  for (i=0;i<7;i++)
    if (PL(x+dirsx[i],y+dirsy[i]) == pl) return 1;
  return 0;
}


void dosave (char *name)
{
  FILE *fp;
  int i,j,k,x,y;

  fp = fopen (name,"w");
  fprintf (fp,"#! atlantis -lo\n");
  fprintf (fp,"%d %d %d %d\n",nplay,pl,hdx,hdy);
  for(ALLSECTORS (i,j)) {
    if (!hassunk (i,j)) {
      fprintf (fp,"%d %d ",i,j);
      for (ALLDIRS(k)) {
	x = SECTX(i,j)+dirsx[k];
	y = SECTY(i,j)+dirsy[k];
	if (ISOFFBOARD (x,y))
	  fprintf (fp,"%c ", CRATER);
	else if (ISEMPTY (x,y))
	  fprintf (fp,"%c ", EMPTY);
	else
	  fprintf (fp,"%c%d%d ",ISBRON(x,y)?SOURCE:PLAYER,
		   PL(x,y),NP(x,y));
      }
      fprintf (fp,"\n");
    }
  }
  fclose (fp);
}

int getnb (FILE *fp)
{
  int ch;

  while (((ch = getc (fp)) == ' ') ||
	 (ch == '\t') ||
	 (ch == '\r') ||
	 (ch == '\n'))
    /* Do nothing */;
  return (ch);
}


void doload (char *name)
{
  char buf[80];
  FILE *fp;
  int i,j,k,x,y,ch;

  clearboard ();
  fp = fopen (name,"r");
  fgets (buf,80,fp);
  fscanf (fp,"%d %d %d %d",&nplay,&pl,&hdx,&hdy);
  while (fscanf (fp,"%d %d",&i,&j) == 2) {
    x = SECTX (i,j);
    y = SECTY (i,j);
    sectors [i][j] = -1;
    for (ALLDIRS(k)) {
      ch = getnb (fp);
      switch (ch) {
      case CRATER:MAKEKRATER (x+dirsx[k],y+dirsy[k]);break;
      case EMPTY:SETONBOARD (x+dirsx[k],y+dirsy[k]);break;
      case SOURCE:
      case PLAYER:SETONBOARD (x+dirsx[k],y+dirsy[k]);
	SETPLAYER (x+dirsx[k],y+dirsy[k],getc (fp)-'0');
	SETPIECES (x+dirsx[k],y+dirsy[k],getc (fp)-'0');
	if (ch == SOURCE)
	  SETBRON (x+dirsx[k],y+dirsy[k]);
	break;
      default:
	printf ("Illegal save file format.\n");
	return;
      }
    }
  } 
  fclose (fp);
  /* calcscreentrans (); */
  /* doredraw (); */
}


#define PUSH(x_,y_) {   traceback[  sp].x=x_;   traceback[sp++].y=y_;}
#define POP(x_,y_)  {x_=traceback[--sp].x   ;y_=traceback[sp  ].y;   }


int floodfill (int x,int y,int *pl)
{
  struct xy {int x,y;};

  struct xy *traceback;
  int max=MAXB*MAXB;
  int sp;
  int cnt = 0;
  int i,xx,yy;

  traceback = alloca (max*sizeof (struct xy));

  *pl = -2;
  sp =0;
  PUSH (x,y);

  while (sp > 0) {
    POP (x,y);
    if (ISFILLED (x,y)) continue;
    SETFILLED (x,y);
    cnt++;
    if (NP(x,y) != 0) {
      if (PL(x,y) != *pl) {
	if (*pl == -2) *pl = PL (x,y);
	else       *pl = -1;
      }
    }
    for (i=1;i<7;i++) {
      xx = x + dirsx[i];
      yy = y + dirsy[i];
      if ((xx > 0) && (yy > 0)) {
	if(ISBOARD (xx,yy) && !ISFILLED (xx,yy))
	  PUSH (xx,yy);
	if(ISBRON (xx,yy))
	  if (PL(x,y) != *pl)
	    if (*pl == -2) *pl = PL (x,y);
	    else           *pl = -1;
      }
    }
  }
  return cnt;
}





int checkscore ()
{
  int curcnt;
  int i;
  int x,y,pl;

  for (i=0;i<MAXPL;i++)
    score[i] = 0;

  for(ALLHEXS(x,y))
    CLRFILLED(x,y);

  /* while (!done) */
  for(ALLHEXS(x,y))
    {
      if (!ISFILLED (x,y))
        {
	  curcnt = floodfill (x,y,&pl);
	  score[pl+2] += curcnt;
        }
    }
  for (i=0;i<nplay+2;i++)
    printf ("Score[%d] = %d\n",i-2,score[i]); 
  return (score[1]);
}



int main (int argc,char **argv)
{
  int np = 6,i,j,sz= 6;
  FILE *fp = NULL;
  int sckt;
  int port = 10000;

  for (i=1;i<argc-1;i++)
    if (strcmp (argv[i],"-np") == 0) {
      np = atoi (argv[i+1]);
      argc-= 2;
      for (j=i;j<argc;j++)
	argv[j] = argv[j+2];
      i -= 2;
    } else if (strcmp (argv[i],"-si") == 0) {
      sz = atoi (argv[i+1]);
      argc-= 2;
      for (j=i;j<argc;j++)
	argv[j] = argv[j+2];
      i -= 2;
    } else if (strcmp (argv[i],"-bo") == 0) {
      fp = fopen (argv[i+1],"r");
      if (fp == NULL) {
	fprintf (stderr,"Can't open '%s':",argv[2]);
	perror ("");
	exit (1);
      }
      argc-= 2;
      for (j=i;j<argc;j++)
	argv[j] = argv[j+2];
      i -= 2;
    } else if (strcmp (argv[i], "-po") == 0) {
      port = atoi (argv[i+1]);
      argc-= 2;
      for (j=i;j<argc;j++)
	argv[j] = argv[j+2];
      i -= 2;
    }
  initboard (np,sz,fp);
  printf ("Serving a board of %d size, with %d players.\n",sz,np);
  sckt = initsocket (port);
  init_undo ();
  serve (sckt);
  exit (0);
}

