#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>


#include "atlantis.h"
#include "serve.h"
#include "commands.h"
#include "undo.h"
#include "ral.h"

int allready = 0;


struct user_str {
  char *name;
  char *passwd;
  struct clientstr *client;
} users[MAXPL] = {{NULL,NULL},};

int maxused_x;                        /* PK: highest coords ever used */
int maxused_y;

int help (int argc,char **argv)
{
  int i;
  int ok=0;

  if (argc == 1) {
    for (i=0;i<nbuiltins;i++)
      cprintf ("250-%12s  %s\n",
	       builtins[i].name,builtins[i].synopsis);
    cprintf ("250 Use help <command> to get more help on a command.\n");
  } else{
    for (i=0;i<nbuiltins;i++)
      if (strcmp (argv[1],builtins[i].name) == 0) {
	cprintf ("250 %12s  %s\n",
		 builtins[i].name,builtins[i].help);
	ok++;
      }
    if (!ok)
      cprintf ("450 Command %s not implemented.\n",argv[1]);
  }
  return 1;
}


int params (int argc,char **argv)
{
  cprintf ("110 %d %d %d  - %d players, on a %dx%d board.\n",
	   nplay,maxused_x,maxused_y,
	   nplay,maxused_x,maxused_y);
  return 0;
}


int quit (int argc,char **argv)
{
  if (argc)
    cprintf ("180 Goodbye.\n"); 
  if (client->pln >= 0)
    users[client->pln].client = NULL;
  fclose (curfp);
  ral_close (curfd);
  FD_CLR (curfd,&allfds);
  curst = NO_CLIENT;
  return 0;
}


int getboard (int argc,char **argv)
{
  int i,j,k,x,y;

  cprintf ("120- %d %d %d %d\n",nplay,pl,hdx,hdy);
  for(ALLSECTORS (i,j)) {
    if (!hassunk (i,j)) {
      cbprintf ("121- %d %d",i,j);
      for (k=0;k<7;k++) {
	x = SECTX(i,j)+dirsx[k];
	y = SECTY(i,j)+dirsy[k];
	if (ISOFFBOARD (x,y))
	  cbprintf (" %c", CRATER);
	else if (ISEMPTY (x,y))
	  cbprintf (" %c", EMPTY);
	else
	  cbprintf (" %c%d%d",ISBRON(x,y)?SOURCE:PLAYER,
		    PL(x,y),NP(x,y));
      }
      cprintf ("\n");
    }
  }
  cprintf ("122 That's it.\n");
  return (0);
}


#if 0

int pln (int argc,char **argv)
{
  int n;

  if (argc <= 1) {
    cprintf ("150 %d - your player number is %d.\n",
	     client->pln,client->pln);
    return 0;
  }

  n = atoi (argv[1]);
  if ((n < -1) || (n >= nplay) || getplayer (n)) {
    cprintf ("440 Player number %d not ok. You remain player %d.\n",
	     n,client->pln);
    return 1;
  }

  if (client->st !=  INITIALIZING) {
    cprintf ("455 You're no longer initializing. Can't change player number now.\n");
    return 1;
  }

  client ->pln = n;
  cprintf ("170 %d - new player number %d ok.\n",n,n);
  return (0);
}

#endif

int readytoplay (int argc,char **argv)
{
  int i,nready;

  if (client -> pln < 0) {
    cprintf ("430 You need a valid player number. %d is not valid.\n",
	     client->pln);
    return 1;
  }
  for (i=0;i<MAXCLIENTS;i++) {
    if ((clients[i].st == READYTOPLAY) &&
	(client->pln == clients[i].pln)) {
      cprintf ("420 Your player number %d is already taken. Find another one....\n",client->pln);
      return 1;
    }
  }
  client->st = READYTOPLAY;
  nready=0;
  for (i=0;i<MAXCLIENTS;i++)
    if (clients[i].st == READYTOPLAY)
      nready ++;
  if (allready)
    cprintf ("141 Welcome back...\n");
  else if (nready == nplay) {
    allready = 1;
    aprintf ("141 ALL players are ready, let the games begin!\n");
    for (i=0;i<nplay;i++)
      player[i]= NULL;
    for (i=0;i<MAXCLIENTS;i++)
      if (clients[i].st == READYTOPLAY)
	player[i] = &clients[i];
    pl = 0;
    pprintf (0,"142 It's your turn.\n");
    init_moves (pl);
    add_cp (BRD);
  }
  else
    cprintf ("140- Not all players are ready yet, please wait.\n");
  return 0;
}

int x0,y0;


#define CHECK_MY_TURN()                                               \
    if  (client -> pln != pl) {                                       \
      cprintf ("400 Shut up: its not your turn!\n");                  \
      return 1;                                                       \
    }                                                                 \
    if (!allready) {                                                  \
      cprintf ("460 Shut up: Wait till everybody is ready.\n");       \
      return 1;                                                       \
    }                                                            


int source (int argc,char **argv)
{
  CHECK_MY_TURN ();
  x0 = y0 = -1;
  if ((sscanf (argv[1],"%d",&x0) != 1) ||
      (sscanf (argv[2],"%d",&y0) != 1) ) {
    cprintf ("401 Syntax error.\n");
    return 1;
  }
  printf ("%cBoard [%d,%d] = %04x. %d\n",
	  NPP(x0,y0)?' ':'*',x0,y0,brd[x0][y0],NPP(x0,y0));
  if (NPP(x0,y0) > 0) {
    cprintf ("160 %d Source OK: %d stones allowed to move.\n", 
	     NPP(x0,y0),NPP(x0,y0));
  } else {
    if (PL (x0,y0) != pl)
      cprintf ("402 Not your stones.\n");
    else
      cprintf ("403 No stones at all.\n");
    x0 = y0 = -1;
    return 1;
  }
  return 0;
}


void update_clients (void)
{
  int x,y,i,j,k;


  for(ALLSECTORS(i,j))
    if (hassunk(i,j) && hasmodified(i,j)) {
      printf ("Sector %d,%d has just sunk....\n",i,j);
      clearmodified (i,j);
      aprintf ("129 %d %d - This segment has sunk.\n",i,j);
    }

  for(ALLSECTORS(i,j)) {
    if (hasmodified (i,j)) {
      clearmodified (i,j);
      abprintf ("130- %d %d",i,j);
      for (k=0;k<7;k++) {
	x = SECTX(i,j)+dirsx[k];
	y = SECTY(i,j)+dirsy[k];
	if (ISOFFBOARD (x,y))
	  abprintf (" %c", CRATER);
	else if (ISEMPTY (x,y))
	  abprintf (" %c", EMPTY);
	else
	  abprintf (" %c%d%d",ISBRON(x,y)?SOURCE:PLAYER,
		    PL(x,y),NP(x,y));
      }
      aprintf ("\n");
    }
  }
}


int dest (int argc,char **argv)
{
  int x1,y1,n,d,i;

  CHECK_MY_TURN ();
  if ((x0 < 0) || (y0 < 0)) {
    cprintf ("404 You should set the source first.\n");
    return 1;
  }
  if ((sscanf (argv[1],"%d",&x1) != 1) ||
      (sscanf (argv[2],"%d",&y1) != 1) ) {
    cprintf ("401 Syntax error.\n");
    return 1;
  }
  convert (x0,y0,x1,y1,&n,&d);
  if (d == 0) {
    cprintf ("405 Not straight.\n");
    return 1;
  }
  if (n > NPP (x0,y0)) {
    cprintf ("406 Can't move %d stones %d spaces....\n",NPP(x0,y0),n);
    return 1;
  }
  for (i=0;i<=n;i++)
    if (!ISBOARD (x0+dirsx[d]*i,y0+dirsy[d]*i)) {
      cprintf ("407 Nah. You can't jump over a non-board position.\n");
      return 1;
    }
  /* Seems like a valid move.... :-) */
  domove (x0,y0,n,d);
  add_move (x0,y0,n,d);
  update_clients ();
  return 0;
}



int done (int argc,char **argv)
{
  CHECK_MY_TURN ();
  checkplof (pl);
  dobron (pl);
  pl = (pl+1) % nplay;
  update_clients ();
  init_moves (pl);
  checkscore ();
  add_cp (BRD); 
  dump_state (BRD);
  aprintf ("135- %d %s is now allowed to play.\n",pl,users[pl].name);
  pprintf (pl,"142 It's your turn.\n");
  return 0;
}


int undo (int argc,char **argv)
{
  CHECK_MY_TURN ();
  do_undo ();
  update_clients ();
  return 0;
}


int redo (int argc,char **argv)
{
  CHECK_MY_TURN ();
  do_redo ();
  update_clients ();
  return 0;
}


int user (int argc,char **argv)
{
  char *tmp;
  int i;

  if (argc <= 1) {
    if (client ->pln < 0) {
      cprintf ("161 Viewer: You are not active in the game.\n");
      return (0);
    }
    tmp = users[client->pln].name;
    if (tmp)
      cprintf ("161 %s Your name is %s.\n",tmp,tmp);
    else
      cprintf ("161 NONE . What's your name anyway?\n");
    return (0);
  }
  tmp = malloc (strlen (argv[1])+1);
  strcpy (tmp,argv[1]);

  for (i=0;i<MAXPL;i++) {
    if (users[i].name &&
	(strcmp (users[i].name,tmp) == 0)) {
      if (users[i].client)
	cprintf ("441 Name %s already taken.\n",tmp);
      else {
	users[i].client = client; /* Verify password */
	client->pln = i;
	cprintf ("162 Your name %s is ok.\n",tmp);
	readytoplay (argc,argv);
      }
      return (0);
    }
  }
  if (argc > 2) {
    sscanf (argv[2],"%d",&i);
    if (!users[i].name) {
      users[i].name = tmp;
      users[i].client = client; /* set password? */
      client ->pln = i;
      cprintf ("162 Your name is now %s.\n",tmp);
      readytoplay (argc,argv);
      return (0);
    } else {
      cprintf ("463 Your preferred playernumber is already taken.\n");
      return (1);
    }
  }
  for (i=0;i<MAXPL;i++) {
    if (!users[i].name) {
      users[i].name = tmp;
      client ->pln = i;
      users[i].client = client; /* Set password? */
      cprintf ("162 Your name is now %s.\n",tmp);
      readytoplay (argc,argv);
      return (0);
    }
  }
  printf ("Huh. Error.\n");
  return (0);
}



int getscore (int argc,char **argv)
{
  char buf[100];
  int bp=0;
  int i;

  for (i=0;i<nplay+2;i++)
    bp += sprintf (buf+bp," %d",score[i]);
  cprintf ("165%s\n",buf);
  return (0);
}


struct clientstr *getclientbyname (char *name)
{
  int i;

  for (i=0;i<MAXPL;i++) {
    if (users[i].name &&
	(strcmp (users[i].name,name) == 0))
      return users[i].client;
  }
  return NULL;
}


int whois (int argc,char **argv)
{
  int pln;

  if (argc < 2) {
    cprintf ("445 Whois who?\n");
    return 1;
  }

  pln = atoi (argv[1]);
  if (!users[pln].client) {
    cprintf ("300 %d isn't logged in yet\n",pln);
    return (0);
  }
  if (users[pln].name) {
    cprintf ("168 %d %s. (player #%d is called '%s')\n",
	     pln, users[pln].name, pln, users[pln].name);
    return (0);
  }
  cprintf ("300 %d didn't register a name. That's a bug.\n", pln);
  printf ("player %d didn't register a name. That's a bug.\n", pln);
  return 2;
}


int tell (int argc,char **argv)
{
  int i,bp=0;
  struct clientstr *cl;
  char buf[100];
  FILE *tempfp;

  if (argc < 3) {
    cprintf ("445 Tell who what?\n");
    return 1;
  }
  if (!(cl = getclientbyname (argv[1]))) {
    cprintf ("446 Who is '%s' anyway?\n",argv[1]);
    return (1);
  }
  if (client->pln < 0) {
    cprintf ("447 Who are you anyway?\n");
    return (1);
  }

  for (i=2;i<argc;i++)
    bp += sprintf (buf+bp," %s",argv[i]);

  tempfp = curfp;
  curfp = cl->fp;
  cprintf ("178 %s:%s\n",users[client->pln].name,buf);
  curfp = tempfp;
  cprintf ("179 Ok.\n");
  return (0);
}


int say (int argc,char **argv)
{
  int i,bp=0;
  char buf[100];

  if (argc < 2) {
    cprintf ("446 Say what?\n");
    return 1;
  }

  for (i=1;i<argc;i++)
    bp += sprintf (buf+bp," %s",argv[i]);
  aprintf ("178 %s:%s\n",users[client->pln].name,buf);
  return (0);
}



