
/*  
 *  Declare functions that implement commands and 
 *  define the table that contains them.
 *  
 *  To declare the functions simply include this file.
 *  
 *  One source file should define the table:
 *  First include this file to declare the functions 
 *  then #define NEED_TABLE, and include this file again.
 *  
 */

#ifndef NEED_TABLE
#define D(a,b,c,d) int a (int argc,char **argv);
#else
#define D(a,b,c,d) {a,b,c,d},
#endif



#ifndef NEED_TABLE

struct builtinfn {
int (*func)();
char *name;
char *synopsis;
char *help;
};
#endif

#ifdef NEED_TABLE
/* Table preamble: */

struct builtinfn builtins[] = {
#else
extern struct builtinfn builtins[];
#endif




D(help,        "help",       "help [command]",
			"gives help text on commands")

D(params,      "params",     "params",
			"prints the params of the current game")

D(getboard,    "getboard",   "getboard",
			"get the current status of the game")

#if 0
/* these ahve been superseeded by "user" */
D(pln,         "pln",        "pln [number]",
			"get or set the player-number")
D(readytoplay, "readytoplay","readytoplay",
			"indicate that preliminary stuff "
			"is over, and let the games begin....")
#endif

D(user,        "user",       "user [name] [preferred playerno]",
				"set username and enter the game.")

D(source,      "source",     "source x y",
			"Indicate the source hex of a move")

D(dest,        "dest",       "dest x y",
			"Indicate the destination hex of a move")

D(done,        "done",       "done",
			"Indicate that the current players move is over")
D(undo,        "undo",       "undo",
				"undo the last move")

D(redo,        "redo",       "redo",
				"redo the last move that was undone")

D(getscore,    "getscore",   "getscore",
				"Get a list of the score so far")

D(whois,       "whois",      "whois pln",
				"Get the username of player pln")

D(tell,        "tell",       "tell playername string",
				"tell player with name playername the string")

D(say,         "say",        "say string",
				"says string to all other players")

D(quit,        "quit",       "quit",
				"quit the game")


#ifdef NEED_TABLE
/* Table postamble */
};
#else
extern const int nbuiltins;
#endif

#ifdef NEED_TABLE
const int nbuiltins = (sizeof (builtins)/sizeof (struct builtinfn));
#endif

#undef D

