#include <stdio.h>
#include <stdlib.h>


#include "atlantis.h"
#include "ui.h"
#include "undo.h"

typedef struct undo_str *undo;

enum type_t {INVALID,MOVE,CHECK};
#define FAST_REFRESH

struct undo_str {
undo prev,next;
enum type_t type;
int x,y,n,d,pl;
int *board;
};



undo first,current;



undo new_undo (void)
{
  undo tmp;

  tmp = malloc (sizeof (struct undo_str));
  tmp->type = INVALID;
  tmp->x = tmp->y = tmp->n = tmp->d = 0;
  tmp->prev = tmp->next = NULL;
  tmp->board = NULL;
  return (tmp);
}

void init_undo (void )
{
  current = first = new_undo ();
}

void freelist (undo p)
{
  undo t;

  for (;p != NULL;p=t) {
    t = p->next;
    if (p->board) free (p->board);
    free (p);
  }
}


void insert_undo (undo p)
{
  if (current->next != NULL)
    freelist (current->next);
  p->next = NULL;
  p->prev = current;
  current->next = p;
  current = p;
}


void add_move (int x,int y,int n, int d)
{
  undo p;

  printf ("adding move %d,%d (%d) %d \n",x,y,n,d);
  p = new_undo ();
  p->type = MOVE;
  p->x = x;
  p->y = y;
  p->n = n;
  p->d = d;
  p->pl = pl;
  insert_undo (p);
}

void add_cp (void *brd)
{
  undo p;

  p = new_undo ();
  p->type = CHECK;
  p->board = malloc (BRD_SIZE);
  memcpy (p->board ,brd,BRD_SIZE);
  p->pl = pl;
  insert_undo (p);
}


void dump_state (void *brd)
{
  undo p;

  if (current->prev == NULL) return;

  printf ("In dump_state...\n");

  for (p=current->prev;(p != NULL) && (p->type != CHECK);p=p->prev)
    /* Nothing */;
  if (p == NULL) return;

  for (;p != current;p=p->next)
    if (p->type != CHECK) {
      printf ("### source %d %d\n",p->x,p->y);
      printf ("### dest   %d %d\n",p->x+p->n*dirsx[p->d],
	      p->y+p->n*dirsy[p->d]);
    }
  printf ("### done\n");
}



void do_undo (void)
{
  static int tbrd[MAXB][MAXB];
  int i,j;
  undo p;

  if (current->prev == NULL) return;

  printf ("In do_undo...\n");
  memcpy (tbrd ,brd,BRD_SIZE); 

  if (current->type == CHECK) {
    printf ("Can't undo more ...\n");
    return; 
  }

  for (p=current->prev;(p != NULL) && (p->type != CHECK);p=p->prev)
    /* Nothing */;
  if (p == NULL) return;
  memcpy (brd,p->board,BRD_SIZE);

  pl = p->pl;
  for (;p != current;p=p->next)
    if (p->type != CHECK) 
      domove (p->x,p->y,p->n,p->d);
  for (i= 0;i<MAXB;i++)
    for (j=0;j<MAXB;j++)
      if ((NP (i,j) != tNP (i,j)) ||
	  (PL (i,j) != tPL (i,j)))
	SETMODIFIED (i,j);
  current = current->prev;
}


void do_redo (void)
{
  if (current->next == NULL) 
    return;
  current = current->next;
  pl = current->pl;
  switch (current->type) {
  case CHECK:
    memcpy (brd,current->board,BRD_SIZE);
    /* doredraw (); */
    break;
  case MOVE:
    domove (current->x,current->y,current->n,current->d);
    /* dopredraw (); */
    break;
  case INVALID:
  default:
    printf ("Error in redo: Invalid type.\n");
    return;
  }
} 

void dump_undos (void)
{
  undo p;

  printf ("First should be invalid (0)\n");
  for (p=first;p!=NULL;p=p->next) {
    if (p == current)
      printf ("This is the current record:\n");
    switch (p->type) {
    case CHECK:
      printf ("Checkpoint pl = %d.\n",p->pl);
      break;
    case MOVE:
      printf ("Moving vrom %d,%d with %d stones in direction #%d (pl=%d).\n",
	      p->x,p->y,p->n,p->d,p->pl);
      break;
    case INVALID:
    default:
      printf ("Invalid: (%d).\n",p->type);
      break;
    }
  }
}
	
